def print_bunnies(bunnies):
    s = ""
    for b in bunnies[:-1]:
        s += b + ", "
    if len(bunnies) > 0:
        s += "and " + bunnies[-1]
    return s

class Level:
    def __init__(self,
                 msg_saved,
                 msg_dead):
        self.msg_saved = msg_saved
        self.msg_dead = msg_dead

    def complete(self, bunnies, **kwargs):
        kwargs["bunnies"] = print_bunnies(bunnies)
        if len(bunnies) > 0:
            print(self.msg_saved.format(**kwargs))
        else:
            print(self.msg_dead.format(**kwargs))

level_1 = Level("Well done, {bunnies} survived the transport to the lab.",
                "Oh no, none of the bunnies made it to the lab.")
level_2 = Level("Great job, you saved {bunnies}.",
                "Oh no, this was definitely not the right device to shave bunnies."
                "None of {bunnies} made it.")
level_3 = Level("Great job, you saved {bunnies}.",
                "Oh no, {bunnies} got the wrong food.")
level_4 = Level("Great job, you saved {bunnies}.",
                "Oh no, turns out injecting bleach isn't the best idea.")
level_5 = Level("Great job, you and {bunnies} escape into the night.",
                "Oh no, you entered the wrong code and got caught.")

levels = [level_1, level_2, level_3, level_4, level_5]



import argparse
from bunny_lab.server import HTTPServer, RequestHandler

parser = argparse.ArgumentParser(description='The bunny lab server.')
parser.add_argument('--ip', type=str, help='The IP address to listen to.')
parser.add_argument('--port', type=int, help='The port to listen to')

args = parser.parse_args()
ip = args.ip
port = int(args.port)

#
# Start server
#

print(ip, port)

server = (ip, port)
httpd = HTTPServer(server, RequestHandler)
httpd.serve_forever()



import os
os.environ["BUNNY_LAB_LOCAL"] = "1"
import bunny_lab.api as api
import traceback
import sys
from bunny_lab.levels import levels

class BunnyLabSession:
    """
    Client-side interface to the bunny-lab server.
    """
    def __init__(self, url):
        self.url = url
        self.name = api.register(url)
        self.levels_completed = set()
        print(f"Welcome, {self.name}, to the Bunny Lab.\n"
              "Good luck saving your furry comrades!")

    def __del__(self):
        api.unregister(self.url, self.name)

    def get_results(self):
        response = api.get_results(self.url)
        bold_name = "<b>" + self.name + " (you) </b>"
        return response.replace(self.name, bold_name)

    def bunnies_saved(self, bunnies, level=0, **kwargs):
        if not level in self.levels_completed:
            levels[level - 1].complete(bunnies)
            n_saved = len(bunnies)
            api.bunnies_saved(self.url, self.name, n_saved)
            self.levels_completed.add(level)


session = None

def initialize(url):
    global session
    session = BunnyLabSession(url)

def rescue(bunnies, level=0, **kwargs):
    session.bunnies_saved(bunnies, level, **kwargs)


class BunnyLabLevel:
    def __init__(self, level_number, function):
        self.level_number = level_number
        self.function = function
        self.tries = 0
        self.completed = False
        self.n_rescued = 0

    def __call__(self, *args, **kwargs):
        if self.completed:
            print(f"You already completed this level and rescued {self.n_rescued} bunnies.")
            return None

        if self.tries >= 3:
            print("You exhausted your number of tries.")
            return None

        try:
            global rescue
            def rescue_level(bunnies):
                self.n_rescued += len(bunnies)
                session.bunnies_saved(bunnies, self.level_number)
            rescue = rescue_level
            self.function(*args, **kwargs)
            self.completed = True
        except Exception as e:
            self.tries += 1
            print(f"Oh no, your function call failed! You have {3 - self.tries} tries left.\n")
            traceback.print_exc(file=sys.stdout, limit=-1)


#
# Level 1
#

def _transport_bunnies(velocity):
    """
    Bunnies need to be transported from the farm to the lab but the road
    is dangerous and the amount of oxygen in the truck is limited.
    """
    bunnies = ["Fifi", "Gunter", "Hazy"]
    distance = 100.0

    time_required = distance / velocity
    oxygen_required = time_required * 10.0

    if velocity > 100:
        return rescue([])
    if oxygen_required > 20.0:
        return rescue([])

    rescue(bunnies)

transport_bunnies = BunnyLabLevel(1, _transport_bunnies)

#
# Level 2
#

def _shave_bunnies(device_index):
    """
    Shaves bunnies to prepare them for experiments.
    """
    bunnies = ["Alfie", "Bugs", "Chip"]
    devices = ["harmless bunny shaver", "buzz saw", "rusty razor"]
    devices.reverse()
    device = devices[device_index]

    for bunny in bunnies:
        if device in ["buzz saw", "rusty razor"]:
            bunnies.remove(bunny)

    rescue(bunnies)

shave_bunnies = BunnyLabLevel(2, _shave_bunnies)

#
# Level 3
#

def _feed_bunnies(names):
    """
    Some of the bunnies are hungry and need food. This
    function feeds a list of bunnies.
    """
    hungry_bunnies = ["Hips", "Ivan", "Jo"]
    allergies = {"Hips":  ["Carrots"],
                 "Ivan": ["Hay, Carrots"],
                 "Jo": ["Cale"]}
    foods = ["Carrots", "Hay", "Cale"]

    bunnies_fed = []
    for bunny, food in zip(names, foods):
        if not food in allergies[bunny]:
            bunnies_fed.append(bunny)

    healthy_bunnies = [b for b in bunnies_fed if b in hungry_bunnies]
    rescue(healthy_bunnies)

feed_bunnies = BunnyLabLevel(3, _feed_bunnies)

#
# Level 4
#
from random import shuffle

def _test_drug(choose_drug):
    """
    You offer the scientist help with the experiments in order to save your
    bunny friends. The task they give you is to choose the drug to give
    to each patient.
    """
    bunnies = ["Turbo", "Uma", "Velvet", "Whoopie"]
    drugs = ["Placebo", "Bleach"]

    healthy_bunnies = []
    for bunny in bunnies:
        drugs_shuffled = shuffle(drugs)
        drug = choose_drug(drugs_shuffled)
        if drug == "Placebo":
            healthy_bunnies.append(bunny)

    rescue(healthy_bunnies)

test_drug = BunnyLabLevel(4, _test_drug)

#
# Level 5
#
from random import random

def _free_bunnies(spy_class):
    """
    Just as you and a group of other bunnies want to escape from the lab, you
    discover that the back door is locked with a secret code. To open it, you
    need to spy on a guard who's going out for a smoke.
    """
    bunnies = ["Turbo", "Uma", "Velvet", "Whoopie"]
    drugs = ["Placebo", "Bleach"]

    spy = spy_class()

    secret_code = []
    for i in range(10):
        x = random()
        spy.listen(x)
        secret_code.append(x)

    if secret_code == spy.guess_code():
        rescue(bunnies)

free_bunnies = BunnyLabLevel(5, _free_bunnies)



from bunny_lab.names import get_random_name

HTML_HEADER = """
<!doctype html>

<html lang="en">
<head>
  <meta charset="utf-8">

  <title>Bunny lab results</title>

  <link rel="stylesheet" href="css/styles.css?v=1.0">

</head>

<body>
"""

HTML_FOOTER = """
</body>
</html>
"""


class BunnyLab:

    def __init__(self):
        self.results = {}

    def register_user(self):
        name = get_random_name()
        while (name in self.results.keys()):
            name = get_random_name()

        self.results[name] = 0
        return name

    def unregister_user(self, name):
        print("deregistered", name)
        self.results.pop(name)

    def bunnies_saved(self, user, number):
        self.results[user] += number
        return self.results[user]

    def bunny_saved(self, user):
        self.results[user] += 1
        return self.results[user]

    def print_results(self, full_html=True):
        result = ""
        if full_html:
            result += HTML_HEADER

        result += "<table>\n"
        result += "<th> Player name </th> <th> Saved bunnies </th>"
        key_list = list(self.results.keys())
        keys_sorted = sorted(key_list, key=lambda k: self.results[k], reverse=True)
        for k in keys_sorted:
            result += f"<tr><td>{k}</td><td>{self.results[k]}</td></tr>"
        result += "</table>"

        if full_html:
            result += HTML_FOOTER

        return result




import numpy as np
import os
import random

current_dir = os.path.dirname(__file__)
path = os.path.join(current_dir, "bunny_names.txt")
names = open(path).read().split(",")
names = [n.strip() for n in names]

def get_random_name():
    """
    Returns a random bunny name.
    """
    index = random.randint(0, len(names))
    return names[index]





import os
import urllib
import urllib.parse
import urllib.request

def register(url):
    data = urllib.parse.urlencode({'action': "register"})
    data = data.encode('ascii')
    r = urllib.request.urlopen(url, data)
    return r.msg

def unregister(url, name):
    data = urllib.parse.urlencode({'action': "unregister",
                                   'name': name})
    data = data.encode('ascii')
    r = urllib.request.urlopen(url, data)
    return r.msg

def bunnies_saved(url, name, number):
    data = urllib.parse.urlencode({'action': "saved",
                                   'name': name,
                                   'number': number})
    data = data.encode('ascii')
    r = urllib.request.urlopen(url, data)
    return int(r.msg)

def bunny_saved(url, name, number):
    data = urllib.parse.urlencode({'action': "saved",
                                   'name': name})
    data = data.encode('ascii')
    r = urllib.request.urlopen(url, data)
    return int(r.msg)

def get_results(url):
    r = urllib.request.urlopen(url + "?notebook=1")
    return r.read().decode("utf-8")



from http.server import BaseHTTPRequestHandler, HTTPServer
import urllib.parse
import os
from urllib.parse import urlparse, parse_qs
from bunny_lab.logic import BunnyLab

BUNNY_LAB_DIR = os.path.dirname(__file__)

bunny_lab = BunnyLab()


def handle_reset():
    bunny_lab = BunnyLab()
    return ""

def handle_register_user():
    name = bunny_lab.register_user()
    return name

def handle_unregister_user(name):
    bunny_lab.unregister_user(name)
    return ""

def handle_bunnies_saved(name, number):
    saved_bunnies = bunny_lab.bunnies_saved(name, number)
    return str(saved_bunnies)

def handle_bunny_saved(name):
    saved_bunnies = bunny_lab.bunny_saved(name, number)
    return str(saved_bunnies)

def handle_file_request(path):
    path = os.path.join(BUNNY_LAB_DIR, "static", path)
    return open(path).read()

class RequestHandler(BaseHTTPRequestHandler):

    def confirm_post(self):
        self.send_response(200)
        self.send_header('Content-type', 'text/html')
        self.end_headers()

    def try_serve_file(self):
        parsed_url = urlparse(self.requestline)
        path = parsed_url.path
        try:
            path = os.path.join(BUNNY_LAB_DIR, "static") + path
            response = open(path).read()
            print("READING FILE: ", open(path).read())
            self.send_response(200)
            self.send_header('Content-type', 'text/html')
            self.end_headers()
            results = bunny_lab.print_results()
            self.wfile.write(response.encode())
        except:
            print("FAILED READING FILE: ", path)
            self.send_response(400)
            self.end_headers()

    def generate_and_send_results(self):
        parsed_url = urlparse(self.requestline)
        path = parsed_url.path
        qs = parse_qs(path)
        print(path, qs)
        if "notebook" in qs:
            full_html = False
        else:
            full_html = True

        self.send_response(200)
        self.send_header('Content-type', 'text/html')
        self.end_headers()
        results = bunny_lab.print_results(full_html)
        self.wfile.write(results.encode())


    def do_GET(self):
        self.protocol_version = "HTTP/1.1"

        if self.path[:2] != "/?"[:len(self.path)]:
            self.try_serve_file()
        else:
            self.generate_and_send_results()


    def do_POST(self):
        content_length = int(self.headers['Content-Length'])
        post_data = self.rfile.read(content_length).decode('utf-8')
        post_data = parse_qs(post_data)
        print("Incoming post: ", post_data)
        action = post_data["action"][0]

        if action == "reset":
            name = post_data["name"][0]
            if name == "b.diddy":
                handle_reset()
            message = ""
        elif action == "register":
            message = handle_register_user()
        elif action == "unregister":
            name = post_data["name"][0]
            message = handle_unregister_user(name)
        elif action == "saved":
            name = post_data["name"][0]
            number = int(post_data["number"][0])
            message = handle_bunnies_saved(name, number)
        elif action == "bunny_saved":
            name = post_data["name"][0]
            message = handle_bunny_saved(name)

        self.send_response(200, message)
        self.send_header('Content-type', 'text/html')
        self.end_headers()





import setuptools

setuptools.setup(
    name="bunny_lab",
    version="0.0.1",
    author="Simon Pfreundschuh",
    author_email="simon.pfreundschuh@chalmers.se",
    description="An interactive Python quest to save the bunnies.",
    url="https://github.com/simonpf/bunny_lab",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
    ],
    python_requires='>=3.6',
)
